//
// WishInterpreter.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Class to interpret the users grouping desires
//
#include "CalServer.h"
#include "Dal.h"

#include "WishInterpreter.h"
#include "ExceptionErrorHandler.h"
#include "DatafileAttributes.h"


#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"
#include "stringutil.h"

/* Method: WishInterpreter() - Constructor */
WishInterpreter::WishInterpreter()
{
///  PROC;

  /* Get the units for any user input */
   _units = stringParameter("units");

  /* Get any energy specific information from the response file
     if energy response or spectral resolution oversampling has 
     been requested */
   if (booleanParameter("withrmfset")) {

      // Open the RMF and read the energy of each channel
       string rmfile= stringParameter("rmfset");
       DataSet * rmfset = dataSetServer->open( rmfile, DataSet::Read );
       Table * ebt = rmfset->table("EBOUNDS");

       //Column *chncol = ebt->column("CHANNEL");
       //ColumnData *chndata = chncol->data();

      // Handle the case of int16 and int32 channel data 
       vector<int> echandat;
       
       CalOalUtility::fillVector(ebt, "CHANNEL", echandat);
      // if (chncol->dataType() == Column::Int16) 
        //  int16 *echandat = chndata->int16Data();
     //  else
       //   int32 *echandat = chndata->int32Data();

       Column *emincol = ebt->column("E_MIN");
       ColumnData *emindata = emincol->data();
       real32 *emin = emindata->real32Data();
       Column *emaxcol = ebt->column("E_MAX");
       ColumnData *emaxdata = emaxcol->data();
       real32 *emax = emaxdata->real32Data();
 
       for (unsigned i=0; i<ebt->rows(); ++i) {
           _echan.push_back(echandat[i]);
           _chan_energy.push_back( (emax[i]+emin[i])/2.0 );
       }

       //delete chndata; 
       delete emindata; delete emaxdata;

  // If not - then make sure that the RMF is not needed
   } else {

       if ( (_units == "KEV" && ( booleanParameter("withranges") || 
          booleanParameter("withRegularBins") )) ||
         (booleanParameter("withoversampling") ) ) 

    /* Units of energy have been requested so we need to find the conversion
       between channels and energy */
         errstr<<"Channel energy information needed but no response " 
               <<"matrix supplied"<<Sas::error("NoRMFsupplied");

   }

  // Get the direction for statistical grouping
   _high2Low = booleanParameter("hightolow");

}

/*
  Method: copyFromTemplate
     Purpose: copy the grouping from another file
     Inputs: channels, source counts, background counts, area normalisation 
*/
void WishInterpreter::copyFromTemplate(Spectrum srcspec)
{

  // Set the CAL status with the instrument and mode/order
    SetCALfromSpec(srcspec.srcfile);

  // Use the same groups as defined in an input spectrum
    string grptemplate= stringParameter("grouptemplate");
    DataSet * grptmpset = dataSetServer->open( grptemplate, DataSet::Read );
    Table * spectrumtable = grptmpset->table("SPECTRUM");

   // Ensure that the template spectrum has the same number of channels
   // as the input spectrum.
    if (spectrumtable->rows() != srcspec.channels.size()) 
       errstr<<"The template spectrum must contain the same number "
             <<"of channels as the input spectrum"
             <<Sas::error("IncompatibleTemplate");

   // Read the grouping info and convert into output groups
    //Column *GCol = spectrumtable->column("GROUPING");
    //ColumnData * GColData   = GCol->data();
    //int16      * GrpData   = GColData->int16Data();
    vector<int16> GrpData;

    CalOalUtility::fillVector(spectrumtable, "GROUPING", GrpData);

    _groups = convertGroupVector(GrpData);

   // Set bad channels as requested  - NOT SURE WE WANT TO DO THIS
    // Initialise the UsedSoFar and Quality arrays
    for (unsigned i=0; i<srcspec.counts.size(); ++i) {
              _UsedSoFar.push_back(false);
              _quality.push_back(GOOD);
    }

    setBadChans(srcspec);

}

/*
  Method: interpretWish
     Purpose: interpret the command line parameters and make the groups
     Inputs: channels, source counts, background counts, area normalisation 
*/
void WishInterpreter::interpretWish(Spectrum srcspec)
{

  // Set the CAL status with the instrument and mode/order
    SetCALfromSpec(srcspec.srcfile);

  // Initialise the UsedSoFar and Quality arrays
    for (unsigned i=0; i<srcspec.counts.size(); ++i) {
              _UsedSoFar.push_back(false);
              _quality.push_back(GOOD);
    }

  /* Set the grouping ranges using the following order 
     1. User defined ranges
     2. Regular user defined bins
     3. Set bad, anything defined as bad
     
     Then with whatever is left, if set, use one of the following

     a) Minimum no of counts
     b) Min S/N
     c) Min ratio above background

     These convolved with the options

     i) Minimum bin width as fraction of energy resolution
     ii) Start from low or high energy end
     iii) Treatment of last bin


     NB: Groups are constructed in any order and are sorted at the end 
  */

   // Set ranges
    if (booleanParameter("withranges")) {
       string rangeString = stringParameter("ranges");
       setRanges(srcspec.channels, rangeString);
       errstr<<"Ranges: "<<_groups[0].first<<" "<<_groups[0].second
             << Sas::message(Msg::App, Msg::Verbose);
    }

  // Regular binning
    if (booleanParameter("withRegularBins")) {
       int rbin_start = intParameter("regbinstart") - 1; // bin 1 = bin 0
       int rbin_end = intParameter("regbinend") - 1; // bin 1 = bin 0
       int rbin_width = intParameter("regbinwid");
       regularBins(rbin_start, rbin_end, rbin_width);
    }

  // Set BAD channels before doing any statistical grouping
    setBadChans(srcspec);

  // Check that only one statistical method has been chosen
    int nstat=0;
    if (booleanParameter("withCounts"))
       nstat++;
    if (booleanParameter("withminSN"))
       nstat++;
    if (booleanParameter("withratioabovebgnd"))
       nstat++;
    if (nstat > 1) 
       errstr<<"Please confine the grouping selection to only one statistical method"<<Sas::error("MoreThanOneStatMethod");

  // Bin with minimum number of counts
    if (booleanParameter("withCounts")) 
       statGroup(srcspec.counts, srcspec.bcounts, srcspec.areanorm, 
                     double(intParameter("mincounts")), "withCounts");

  // Minimum signal-to-noise
    if (booleanParameter("withminSN")) 
       statGroup(srcspec.counts, srcspec.bcounts, srcspec.areanorm,
               realParameter("minSN"), "withminSN");

  // Minimum ratio above background
    if (booleanParameter("withratioabovebgnd"))
       statGroup(srcspec.counts, srcspec.bcounts, srcspec.areanorm,
                     realParameter("ratioabovebgnd"), "withratioabovebgnd");
  
  /* If no grouping options have been selected then group into single
     bins using a mincounts=0 option */
    if (nstat==0 && !booleanParameter("withranges") && 
                    !booleanParameter("withRegularBins"))
       statGroup(srcspec.counts, srcspec.bcounts, srcspec.areanorm,
                     0.0, "withCounts");

  // Treat bins which have so far not been grouped
    handleUngrouped();

  // Apply a minimum number of groups per resolution element
    if (booleanParameter("withoversampling")) 
       resolveOversampling(realParameter("oversample"));

}

/* setRanges: Uses simple ranges provided by the user */
void WishInterpreter::setRanges(vector<int> channels, string & rangeString)
{
   /* Check that the range string is valid */
     if (! checkRangeString(rangeString))
         errstr<<"Invalid ranges string - enter string of style '1:10'"
            <<Sas::error("InvalidRangeString");
        
   /* Interpret the input ranges string and write directly into the groups */
     _groups = convertRangeString(channels, rangeString);

   /* Flag the channels which have been used */
     updateUsed();
}

/* Method: checkRangeString
   Purpose: Check the format of the range string
*/
bool WishInterpreter::checkRangeString(string & rangeString)
{
     bool stringok=false;

     for ( unsigned ic=0 ; ic < rangeString.length(); ic++ )
     {
      // Check for the ':' or '-' separator 
       if ( rangeString.find(":") !=string::npos ||
                     rangeString.find("-") !=string::npos )
               { stringok=true; }
     }

     return stringok;
}

/* convertRangeString: Converts a string containg ranges to groups
      Input:  1:10, 11:20 for example
*/
vector<WishInterpreter::Group> WishInterpreter::convertRangeString
                           (vector<int> channels, string & rangeString)
{
     vector<Group> rangeGroups;

   // Loop over each character in the string
     int cstart=0;
     string startStr, endStr;

     for ( unsigned ic=0 ; ic < rangeString.length(); ic++ )
     {
       // Range separator

        if (rangeString.substr(ic,1) == ":" || 
                     rangeString.substr(ic,1) == "-") {
           startStr = rangeString.substr(cstart,(ic-cstart+1));
           cstart=ic+1;
        }
       // Ranges seperator
        if (rangeString.substr(ic,1) == "," || ic == (rangeString.length()-1)) {
           endStr = rangeString.substr(cstart,(ic-cstart+1));
           cstart=ic+1;
        
          // Convert to integer range
           //Group group = make_pair(toNumber<int>(startStr), 
             //                               toNumber<int>(endStr));           
           Group newgroup = groupChanOrEnergy(channels, startStr, endStr);
           rangeGroups.push_back(newgroup);

           errstr<<"Added range: "<<startStr<<" "<<endStr
                 << Sas::message(Msg::App, Msg::Verbose);
        }
     }
     
     return rangeGroups;
}

/* Method: groupChanOrEnergy(startStr, endStr);
   Purpose: Make a group from two strings which may contain  a channel
            or an energy range
*/
WishInterpreter::Group WishInterpreter::groupChanOrEnergy(
                                    vector<int>channels,string &startStr, 
                                                          string &endStr)
{
     Group newgroup;

   /* If units are channels then no need to convert */
     if (_units == "CHAN")
        newgroup = make_pair(toNumber<int>(startStr), toNumber<int>(endStr));

     else if (_units == "KEV") {

       // Convert energy range to channels
        double lowen = toNumber<double>(startStr);
        double highen = toNumber<double>(endStr);

       // Check that low-energy range is not too high and high-energy range
       // is not too low
        if (lowen > max(_chan_energy[_chan_energy.size()-1],_chan_energy[0]) ||
              highen < min(_chan_energy[0],_chan_energy[_chan_energy.size()-1]))

                errstr<<"Energy range not within spectral bounds: "<<
                _chan_energy[0]<<" to "<<_chan_energy[_chan_energy.size()-1]
                <<Sas::error("EnergyOutOfRange");

       // Convert from energy to channels - taking care of direction
        int istart = min(channel_of_energy(lowen), channel_of_energy(highen));
        int iend = max(channel_of_energy(highen), channel_of_energy(lowen));

       // Make the group - converting from channels to spectral bin
        newgroup = make_pair(bin_of_channel(channels, istart), 
                             bin_of_channel(channels, iend));
     }
     return newgroup;
}

/* Method: convertGroupVector(grpVec)
   Purpose: Converts a group vector into output groups
*/
vector<WishInterpreter::Group> WishInterpreter::convertGroupVector
                           (vector<int16> grpVector)
{
     vector<Group> rangeGroups;
     vector<int> start;
     vector<int> stop;

    // Loop over array and find the start channels
     for (unsigned g=0; g<grpVector.size(); g++)
         if (grpVector[g] == 1)
                start.push_back(INT(g));

    // Set the stop channels
     for (unsigned g=0; g<start.size()-1; g++)
         stop.push_back(start[g+1]-1);

     stop.push_back(grpVector.size()-1);

    // Set the output groups
     for (unsigned g=0; g<(start.size()); g++) {
        Group newgroup = make_pair(start[g], stop[g]);
        rangeGroups.push_back(newgroup);
     }

     errstr<<"Found "<<start.size()<<" groups in input template "
           <<start[start.size()-1]<<" "<<stop[start.size()-1]
           << Sas::message(Msg::App, Msg::Verbose);

     return rangeGroups;
}


/* Method: bin_of_channel(int chan);
   Purpose: Find the bin in the spectrum (from 0) for an input channel.
*/
int WishInterpreter::bin_of_channel(vector<int>channels, int inchan)
{
   int bin=0;

  // Set to max if channel number too large
   if (inchan >= channels[channels.size()-1])
      bin=channels.size()-1;

  // Find the bin containing this channel
   for (unsigned i=0; i<channels.size()-1; ++i) 
       if (inchan >= channels[i] && inchan < channels[i+1])
                          bin=i;

   return bin;
}

/* Method: energy_of_channel(int chan);
   Purpose: Find the energy of a given PI channel
*/
double WishInterpreter::energy_of_channel(int chan)
{
     double outen=0.0;

    // Find the energy of this channel
     if (chan >= _echan[_echan.size()-1]) 
        outen = _chan_energy[_echan.size()-1];
     else {
        for (unsigned i=0; i<_echan.size()-1; ++i) {

          if (chan >= _echan[i] && chan < _echan[i+1])
                 outen = _chan_energy[i];
        }
     }

     return outen;
}

/* Method: channel_of_energy(double energy);
   Purpose: Find the channel corresponding to an energy or beta.
*/
int WishInterpreter::channel_of_energy(double energy)
{
     int outchan=0;
     unsigned N=_chan_energy.size()-1;  // Number of last element

    /* Find the energy of this channel taking into account that energy
       may increase or descrese with channel number */
     if (_chan_energy[0] < _chan_energy[N])  // Channel energy increases
     {
        if (energy >= _chan_energy[N])
              outchan=_echan[N];  // Set to the last element
        else if (energy <= _chan_energy[0])
              outchan=_echan[0];  // Set to the first element
        else {
           for (unsigned i=0; i<N; ++i) {
             if (energy >= _chan_energy[i] && energy < _chan_energy[i+1])
               {
                if (abs(energy - _chan_energy[i]) <
                              abs(energy - _chan_energy[i+1]))
                    outchan = _echan[i];
                else
                    outchan = _echan[i+1];
               }
           }
        }
     } else {  // Channel energy decreases

        if (energy >= _chan_energy[0])
              outchan=_echan[0];  // Set to the last element
        else if (energy <= _chan_energy[N])
              outchan=_echan[N];  // Set to the last element
        else {
           for (unsigned i=0; i<N; ++i) {
             if (energy <= _chan_energy[i] && energy > _chan_energy[i+1])
               {
                if (abs(energy - _chan_energy[i]) <
                              abs(energy - _chan_energy[i+1]))
                    outchan = _echan[i];
                else
                    outchan = _echan[i+1];
               }
           }
        }
     }

     return outchan;
}

/* Method: updateUsed
   Purpose: keep a track of the channels which have been used already
*/
void WishInterpreter::updateUsed()
{
   /* This simply sets the flag to used for all of the channels which
      are currently within a group */
     for (unsigned i=0; i<_groups.size(); ++i) {
        for (int j=_groups[i].first; j<=_groups[i].second; ++j)
             { _UsedSoFar[j]=true; }
     }
}

/* regularBins : Uses simple regular binning provided by the user */
void WishInterpreter::regularBins(int rbin_start, int rbin_end, int rbin_width)
{

   // Sanity check
     if (rbin_start < 0) 
        errstr<<"Start bin for regular binning cannot be less than 1"
              <<Sas::error("RegBinRange");

   /* First check that none of the requested bins have already been set */
     for (int i=rbin_start; i<=rbin_end; ++i)
        if (_UsedSoFar[i])
           errstr<<"Ambiguous definition for grouping of spectral bin "
               <<i<<Sas::error("AmbiguousGrouping");

   /* Set regular groups. The last group may be smaller than the required
      width if the requested end-start is not an integer multiple of 
      the width  */
     for (int i=rbin_start; i<=rbin_end; i+=rbin_width) {
         //Group newgroup = new Group(i, min( (i+rbin_width-1),rbin_end ));
         Group newgroup = make_pair (i, min( (i+rbin_width-1),rbin_end ));
         _groups.push_back(newgroup);
     }

   /* Update the used channels array */
     updateUsed();

}

/* setBadChans: Set channels as bad when requested */
void WishInterpreter::setBadChans(Spectrum srcspec)
{
    /* Find the user option selected */
     string setbadoption = stringParameter("setbad");

     vector<Group> badgroup;

     if (setbadoption == "CCF") {

       // Do stuff
        badgroup = getBadRangefromCCF(srcspec);
                    //(startPI - srcspec.specval) / float(srcspec.specdelt));

     } else if (setbadoption == "no") {
       // Do nothing

     } else {  // Must be a range

       /* Check that the range string is valid */
        if (! checkRangeString(setbadoption))
           errstr<<"Invalid ranges string in setbad option - enter "
                 <<"string of style '1:10' or '0.2-0.5'"
                 <<Sas::error("InvalidRangeString");
        
   /* Interpret the input ranges string and write directly into the groups */
        badgroup = convertRangeString(srcspec.channels, setbadoption);
     }

   /* Set these channels bad if they haven't already been grouped */
     for (unsigned i=0; i<badgroup.size(); ++i) {
        for (int j=badgroup[i].first; j<=badgroup[i].second; ++j) {

           // Check if this channel has already been grouped
            if ( _UsedSoFar[j] ) {
               errstr<<"Bin "<<j<<" has already been grouped and will not be "
                     <<"set to bad"<<Sas::warning("AlreadyGrouped");

            } else {
               _quality[j]=BAD;
               _UsedSoFar[j]=true;
            }
        }
     }
}

/* getBadRangefromCCF: get channel range to set bad from CCF */
const vector<WishInterpreter::Group> WishInterpreter::getBadRangefromCCF(
                                         Spectrum srcspec)
{
      vector<Group> badRanges;

    // Set the CAL status with the instrument and mode/order
//      SetCALfromSpec(srcspec.srcfile);

    // Get the server
      SpecQualityServer *ct = 0; ct = calServer.getAtom(ct);

    // Get the bad ranges from the CCF - NB these are stored in PI chans
      short ngroups = ct->nGroups();

    // Can't find a way of translating between LAMBDA/BETA amnd channel
    // for RGS across the CCF and spectrum - we will use the channel
    // numbers for RGS and assume that the spectrum has the default binning

     /* EPIC & RGS / Lambda */
      vector<short> ranges = ct->ranges();

     // Treat RGS as special case because it has Lambda and Beta spectra
      if ( calServer.instrument()->id() == Xmm::RGS1 || 
           calServer.instrument()->id() == Xmm::RGS2 ) { 

        // Read the expected number of channels for a Lambda spectrum
         int nchans = ct->lambda_nchan();

        // If this is a BETA spectrum read in the BETA calibration
         if (srcspec.rgsunits == "rad" ) {  
            ranges = ct->beta_ranges();
            nchans = ct->beta_nchan();
         }

        // Check that spectrum has same number of elements as the default
         if (nchans != INT(srcspec.channels.size()))
            errstr<<"Number of channels in RGS spetrum does not match "
		  <<"the expected calibrated size of "<<nchans
                  <<Sas::error("NonDefaultRGSSpectrum");
      }

     // Loop over each group
      for (int i=0; i<ngroups; ++i) {
           Group newgroup = findspecChanG(ranges[i*2], 
                                        ranges[i*2+1], srcspec);
           badRanges.push_back(newgroup);

           errstr<<"Setting bad channels: "<<ranges[i*2]<<" to "
              <<ranges[i*2+1]<<" "<<newgroup.first<<" "<<newgroup.second
              << Sas::message(Msg::App, Msg::Verbose);
      }
      
      calServer.releaseAtom(ct);

      return badRanges;
}

/* findspecChanG: finds the channels in the spectum corresponding to 
                 a PI channel range */
const WishInterpreter::Group WishInterpreter::findspecChanG(int startPI, 
                                 int endPI, Spectrum srcspec)
{
   /* Calculate the channel which the start and end PI bins fall in */
    int c1 = findspecChan(double(startPI), srcspec);
    int c2 = findspecChan(double(endPI), srcspec);

    return make_pair(c1,c2);
}

/* findspecChan: finds the channel in the spectum corresponding to 
                 a PI value, lambda or beta value */
const int WishInterpreter::findspecChan(double inval, Spectrum srcspec)
{
   /* Calculate the channel which the PI bin falls in */
     int chan = INT(srcspec.specpix + 0.5 + 
                    (inval - srcspec.specval) / srcspec.specdelt);

     //cout<<"findspecChan: "<<inval<<" "<<chan<<endl;

   /* Calculate the bin in the spectrum of this channel */
     return (bin_of_channel(srcspec.channels, chan));
}

/* statGroup: Groups channels so that each group contains 
                   at least the minimum required statistic */
void WishInterpreter::statGroup(vector<int> counts, vector<int> bcounts, 
                             double areanorm, double minStat, 
                             const string function)
{
    /* Loop over each channel. */
     bool groupStarted=false;
     int srcTotal=0;
     int bgdTotal=0;
     int chanstart=0;
     double stat=0.0;
     double src_bgd_Total=0.0;   // Source minus (area normalised) background

    /* We loop over the bins in the direction requested */
     int istart=0, iend=counts.size(), iinc=1;
     if (_high2Low) {istart=(counts.size()-1); iend=-1; iinc=-1;}

     for (int i=istart; i!=iend; i+=iinc) {

        // Is this channel already in a group 
         if (!_UsedSoFar[i]) {   // Ungrouped channel

           // Add these counts to the current group total
            srcTotal += counts[i];
            bgdTotal += bcounts[i];
            src_bgd_Total = srcTotal - bgdTotal*areanorm;

           // Is this a new group
            if (! groupStarted) {   // New group

              // Set the start channel number 
               groupStarted=true;
               chanstart=i;

            }

           // Calculate the statistic for this function
            if (function == "withCounts") 
               stat = double(srcTotal); // Total (src+bgd) counts
            else if (function == "withminSN")
               //stat = src_bgd_Total / sqrt(double(srcTotal)); // S/N
               stat = src_bgd_Total / sqrt( double(srcTotal) + 
                                            bgdTotal*areanorm*areanorm ); // S/N
            else if (function == "withratioabovebgnd")
               stat = src_bgd_Total / (bgdTotal * areanorm); // Ratio above bgd

           // Have we got enough statistic in this group yet ?
            if ( stat >= minStat) {

              // Write a new group - may be going from high energy end
               Group newgroup = make_pair (min(chanstart, i), 
                                           max(chanstart, i));
               _groups.push_back(newgroup);

              // Reset group
               groupStarted=false; srcTotal=0; bgdTotal=0;
            }

        // This channel has been previously grouped (e.g. within a range)
        // This means that any 'in-progress' group didn't make it - so reset
         } else {
            groupStarted=false; srcTotal=0; bgdTotal=0;
         }

    } // end of loop over channels

}

/* Method: handleUngrouped
   Purpose: To set the quality and grouping for all the ungrouped data.
*/
void WishInterpreter::handleUngrouped()
{

   // First of all update the used array
     updateUsed();

   /* Find the users preference for handling ungrouped channels.
      Options are:
        setbad     - set all channels false and dont group them
        owngroup   - make a new group for each section of ungrouped bins
        addtogroup - add ungrouped channels to the previous group 
   */
     string whatToDo = stringParameter("lastbin");

     bool groupStarted=false;
     int chanstart=0;

   /* Handle any ungrouped bins */
     for (unsigned i=0; i<_UsedSoFar.size(); ++i) {
        if (! _UsedSoFar[i]) {

         // Leave the bins ungrouped and set their quality to bad
           if (whatToDo == "setbad") {
               _quality[i]=BAD;

         // If to be grouped find when a new group has been started
           } else if (whatToDo == "owngroup" || 
                         whatToDo == "addtogroup") {

                if (! groupStarted) {
                   groupStarted=true;
                   chanstart=i;
                }
           }
        }

        if (_UsedSoFar[i] || i==( _UsedSoFar.size()-1 )) {

          // Finish the group if started. NB: Quality is already GOOD
           if (groupStarted) { // This is the end of a new group

               groupStarted=false;

              // Define a new group
               int chanend = i; if (_UsedSoFar[i]) chanend=i-1;

               Group newgroup = make_pair(chanstart, chanend);

              // Add on an 'owngroup' or merge in with other groups
               if (whatToDo == "owngroup")
                  _groups.push_back(newgroup);
               else if (whatToDo == "addtogroup") 
                  mergeGroups(newgroup);

               errstr<<"New grp: "<<newgroup.first<<" "<<newgroup.second
                     << Sas::message(Msg::App, Msg::Verbose);
           }
        }
     } // End of the loop over bins
}

/* Method: mergeGroups(Group newgroup)
   Purpose: To add a group onto one of the existing groups
   Action: Changes the start or end of one of the existing groups
           or makes a new group
*/
void WishInterpreter::mergeGroups(Group newgroup)
{

   /* We look for a group which borders on the new group and then merge
      them into one. Depending on whether we start grouping from low
      or high energy we join to a group above or below */
     int lowergroup=-1, highergroup=-1; // In case we can't group it

    // Start from the low end
    //  if (! _high2Low) {

    // Loop over groups
     for (unsigned i=0; i<_groups.size(); ++i) {

      // Is this a contiguous lower group ?
         if (newgroup.first == _groups[i].second+1) 
                  lowergroup=i;

      // Is this a contiguous higher group ?
         if (newgroup.second == _groups[i].first-1) 
                  highergroup=i;
     }

    /* If we are grouping from low to high energy add onto a lower group
       if available, otherwise add to a higher group. Vice-versa if we
       are grouping from high to low energy. If no contiguous group is
       found then make a separate new group */
     if (lowergroup>-1 && ( !_high2Low || highergroup==-1) )
        _groups[lowergroup].second = newgroup.second; // add to lower group

     else if (highergroup>-1)
         _groups[highergroup].first = newgroup.first; // add to higher group

     else {
         errstr<<"Cant find group to merge with - creating a new group "
               <<"("<<newgroup.first<<":"<<newgroup.second<<")"
               <<Sas::warning("NoMergeGroup");

         _groups.push_back(newgroup);
     }
}

/* Method: mergeGroups(int groupNumber)
   Purpose: To merge an existing group with its neighbour
   Action: Remove this group from the _groups array and merge it
           in as though it were a new group
*/
void WishInterpreter::mergeGroups(unsigned groupNumber)
{

   // Make a temporary group
     Group newgroup = _groups[groupNumber];

   // Remove this group from the pack and shuffle everything up
    //  for (unsigned i=groupNumber+1; i<_groups.size(); ++i) {
      //    _groups[i-1]=_groups[i];

   // Remove the last group
       // erase the 6th element
     _groups.erase(_groups.begin()+groupNumber);

     mergeGroups(newgroup);
}


/* Method: resolveOversampling
   Purpose: Ensures that there are a maximum number of groups per
            energy resolution element
   Action: Merges groups if there are too many
*/
void WishInterpreter::resolveOversampling(float oversample)
{

    /* Read the energy resolution as a function of energy */
      SpecQualityServer *ct = 0; ct = calServer.getAtom(ct);

      vector<float> resolution = ct->resolution();
      vector<float> res_energy = ct->resolution_energy();

      double FWHM = 0.0;

      //cout<<"VECRES: "<<resolution[0]<<" "<<resolution[10]<<" "
         // <<res_energy[0]<<" "<<res_energy[10]<<" "<<res_energy[res_energy.size()-1]<<" "
          //<<resolution[999]<<" "<<resolution[res_energy.size()-1]<<endl;

    /* Loop over each defined group. We go backwards because if a
       group is merged with another then the total number of groups
       is reduced */
      for (int ii=(_groups.size()-1); ii>=0; --ii) {

         unsigned i=ii;

     // Find the centre energy and width of this group
         double egrp_min= energy_of_channel(_groups[i].first);
         double egrp_max= energy_of_channel(_groups[i].second);
         double group_cen_energy = (egrp_max + egrp_min) /2.0;
         double group_width = abs(egrp_max - egrp_min);

     // Loop over each energy to find the closest resolution cal point
         if (group_cen_energy < res_energy[0]) 
            FWHM = resolution[0];
         else if (group_cen_energy >= res_energy[res_energy.size()-1])
            FWHM = resolution[res_energy.size()-1];
         else {
            for (unsigned j=1; j<res_energy.size(); ++j) {

              if (group_cen_energy >= double(res_energy[j-1]) &&
                  group_cen_energy <= double(res_energy[j]) ) {
                
                if (abs(group_cen_energy - double(res_energy[j-1])) <=
                    abs(group_cen_energy - double(res_energy[j])) )
                       FWHM = resolution[j-1];
                else
                       FWHM = resolution[j];
              }
            }
         } // Endof if

        // cout<<"EGRP: "<<i<<" "<<egrp_min<<" "<<_groups[i].first<<" "
          //   <<egrp_max<<" "<<_groups[i].second<<endl;
        // cout<<"RES: "<<group_cen_energy<<" "<<group_width
          //   <<" "<<FWHM<<endl;

     // If grouping finer than required resolution merge with neighbour group 
         if (group_width < (FWHM / oversample)) {

                errstr<<" Resolution of group ["<<_groups[i].first<<":"
                      <<_groups[i].second<<"] too narrow - regrouping..."
                      <<"FWHM: "<<FWHM<<" Oversample: "<<oversample
                      << Sas::message(Msg::App, Msg::Verbose);

                mergeGroups(i);
         }

      } // End of group loop

      calServer.releaseAtom(ct);
}

/* Method: groups()   - return the groups information */
vector<WishInterpreter::Group> WishInterpreter::groups()
{
    return _groups;
}

/* Method: quality()   - return the quality information */
vector<short int> WishInterpreter::quality()
{
    return _quality;
}

/* Method: SetCALfromSpec(specfile) - sets the CAL for the instrument/mode */
void WishInterpreter::SetCALfromSpec(const string & specfile)
{

  /* Open the set */
      DataSet *set = dataSetServer->open( specfile, DataSet::Read );

  /* Read the instrument and pass it to the CAL */
      string instrument = set->attribute("INSTRUME")->asString();

      calServer.setInstrument(Xmm::toInstrument(instrument));

  /* If instrument is EPIC then set the observing mode in the CAL */
      if ( ! (instrument == "RGS1" || instrument == "RGS2") )
      {
        /* Set up the CAL with the instrument and submode */
         DatafileAttributes *specAttributes =
                      new DatafileAttributes(specfile);
         delete specAttributes;

      } else { // RGS 1/2 
 
        /* Find the order and set it as the 'mode_id' to be compatible
	   with the EPIC use of the SPECQUAL file */
         Table * spectrumtable = set->table("SPECTRUM");

        /* Get selection expression */
         int order = 1;  // Assume default order if not given
         if (spectrumtable->hasAttribute("RFLORDER"))
             order = spectrumtable->attribute("RFLORDER")->asInt();

        /* The "Mode" is used to differentiate between different settings
           in the EPIC cameras. In RGS, it is the order rather than the mode
           which is used. However, we have instrument-independent code
           in the CAL and so must use an order-mode fudge. So.....

           If order=1, set mode_offset=0
           If order=2, set mode_offset=1
        */
         //Xmm::Mode pseudo_mode = Xmm::toMode(order - 1);
         Xmm::Mode pseudo_mode = Xmm::toMode(Xmm::SPECTROSCOPY+order-1);

         calServer.state()->mode()->set(pseudo_mode);
      }
}

WishInterpreter::~WishInterpreter()
{
  PROC;
  ENDPROC;
}

